from abc import ABC, abstractmethod

from django_spire.core.notification.email import SendGridEmailHelper


class NotificationSender(ABC):
    def __init__(self, notification: 'Notification'):
        self.notification = notification

    @abstractmethod
    def send(self):
        pass


class EmailNotificationSender(NotificationSender):

    def send(self):
        template_data = {
            'title': self.notification.title,
            'body': self.notification.body,
            'name': self.notification.name,
            'button_url': self.notification.url,
        }
        SendGridEmailHelper(
            to=self.notification.email,
            template_data=template_data,
            fail_silently=False
        ).send()



from django.utils import timezone
from datetime import datetime, timedelta

from django_spire.user_account.factories import get_or_create_user_profile


class UserSettingsMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        """
            Activates the user's timezone and updates the user's settings in the session every 5 minutes.
        """
        if request.user.is_authenticated:
            profile = get_or_create_user_profile(request.user)
            last_update = request.session.get('user_options_expiry_time')

            if last_update is not None:
                last_update = datetime.fromtimestamp(last_update)

            if last_update is None or datetime.now() - last_update >= timedelta(minutes=5):
                request.session['user_options'] = profile.options.to_dict()
                request.session['user_options_expiry_time'] = datetime.now().timestamp()

            timezone.activate(profile.get_option('system', 'timezone'))

        response = self.get_response(request)

        return response



from django import forms

from django_spire.cookbook.recipe.nutrition import models


class NutritionFactForm(forms.ModelForm):
    class Meta:
        model = models.NutritionFact
        exclude = ['recipe']




from django_spire.core.notification.models import Notification


def process_notifications():
    for notification in Notification.objects.filter(is_sent=False):
        notification.send()
    return 'Successfully Completed'



from django.contrib import admin
from django_spire.history import models


class ActivityLogAdmin(admin.ModelAdmin):
    list_display = ('content_object', 'content_type', 'created_datetime', 'verb', 'user')
    list_filter = ('verb',)
    search_fields = ('id', 'actor__first_name', 'user__first_name', 'content_type__model')

    class Meta:
        ordering = ('created_datetime',)


admin.site.register(models.ActivityLog, ActivityLogAdmin)


class EventHistoryAdmin(admin.ModelAdmin):
    list_display = ('content_object', 'created_datetime', 'event_verbose')
    list_filter = ('event',)
    search_fields = ('id', 'content_type__model',)

    class Meta:
        ordering = ('created_datetime',)


admin.site.register(models.EventHistory, EventHistoryAdmin)


class ViewAdmin(admin.ModelAdmin):
    list_display = ('content_object', 'created_datetime', 'user')

    class Meta:
        ordering = ('created_datetime',)


admin.site.register(models.View, ViewAdmin)



from django.db import models

from django_spire.cookbook.recipe.nutrition.enums import NutritionFactType
from django_spire.cookbook.recipe.nutrition.querysets import NutritionFactQuerySet
from django_spire.history.mixins import HistoryModelMixin


class NutritionFact(HistoryModelMixin):
    recipe = models.ForeignKey(
        'cookbook_recipe.Recipe',
        related_name='nutrition_facts',
        related_query_name='nutrition_fact',
        on_delete=models.CASCADE
    )

    type = models.CharField(max_length=255, choices=NutritionFactType.choices)
    quantity = models.FloatField(default=0)

    objects = NutritionFactQuerySet.as_manager()

    def __str__(self):
        return self.type

    def type_verbose(self):
        return NutritionFactType(self.type).label

    class Meta:
        verbose_name = 'Nutrition Fact'
        verbose_name_plural = 'Nutrition Facts'
        db_table = 'recipe_nutrition_fact'



from django.urls import include, path
from django_spire.landing import views

app_name = 'landing'

urlpatterns = [
    path('',
         views.home_view,
         name='home'),
]

# Contact Us #
# urlpatterns += [
#     path('contact-us/',
#          views.contact_us_view,
#          name='contact_us'),
#
#     path('contact-us/success/',
#          views.contact_us_success_view,
#          name='contact_form_success'),
# ]

# Boneyard #
urlpatterns += [
    path('boneyard/',
         views.boneyard_view,
         name='boneyard')
]

# Crypt #
urlpatterns += [
    path('crypt/',
         views.crypt_view,
         name='crypt')
]

# Asylum #
urlpatterns += [
    path('asylum/',
         views.asylum_view,
         name='asylum')
]

# Wasteland #
urlpatterns += [
    path('wasteland/',
         views.wasteland_view,
         name='wasteland')
]

urlpatterns += [
    path('component/',
        include('app.landing.component.urls', namespace='component')
    ),
]



import random
from dateutil import relativedelta

from django.contrib.auth.models import User
from django.contrib.sites.models import Site
from django.db import models
from django.urls import reverse
from django.utils.timezone import localtime

from django_spire.core.notification.models import Notification
from django_spire.core.notification.sender import SendGridEmailHelper
from django_spire.history.mixins import HistoryModelMixin
from django_spire.user_account.authentication.mfa.querysets import MfaCodeQuerySet


class MfaCode(models.Model):
    user = models.ForeignKey(User, on_delete=models.CASCADE, related_query_name='mfa_code', related_name='mfa_codes')
    code = models.CharField(max_length=6, editable=False, unique=True)
    expiration_datetime = models.DateTimeField(editable=False)

    objects = MfaCodeQuerySet.as_manager()

    def __str__(self):
        return f'{self.expiration_datetime} - {self.code}'

    def is_valid(self):
        return self.expiration_datetime > localtime()

    def set_expired(self):
        self.expiration_datetime = localtime()
        self.save()

    def send_notification(self):
        user = self.user
        context_data = {
            'title': 'Authentication Code',
            'body': f"Here is your multifactor authentication code: </br> <h3>{self.code}</h3>",
            'name': user.first_name,
            'button_url': f'{Site.objects.get_current()}/{reverse("user_account:authentication:redirect:login_redirect")[1:]}'
        }

        # Directly connecting to sendgrid to send MFA email faster
        SendGridEmailHelper(
            to=[user.email],
            template_id='d-9ae6be6be95d4f79b0daad9055f03cc9',
            template_data=context_data,
            fail_silently=False
        ).send()

    @classmethod
    def generate_code(cls, user):
        return MfaCode.objects.create(
            user=user,
            code=int(''.join([str(random.randint(0, 9)) for _ in range(6)])),
            expiration_datetime=localtime() + relativedelta.relativedelta(minutes=5)
        )

    class Meta:
        db_table = 'authentication_mfa_code'



import json
from typing import Optional, Union

from django import forms

from django_spire.core.file import widgets
from django_spire.core.file.models import File
from django_spire.core.file.queryset import FileQuerySet


class MultipleFileField(forms.FileField):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.widget = widgets.MultipleWidget()

    def prepare_value(self, value: Optional[list[File]]):
        if value is not None:
            return json.dumps([file.to_dict() for file in value])
        else:
            return json.dumps([])

    def clean(self, data, initial=None) -> dict:
        return data


class SingleFileField(forms.FileField):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.widget = widgets.SingleFileWidget()

    def prepare_value(self, value: Union[File, FileQuerySet, None]) -> Optional[str]:
        if isinstance(value, FileQuerySet):
            value =  value.first()

        if value is not None:
            return value.to_json()
        else:
            return json.dumps(None)

    def clean(self, data, initial=None) -> dict:
        return data



import functools
from typing import Union

from django.core.exceptions import PermissionDenied
from django.http import HttpResponseRedirect
from django.urls import reverse


def permission_required(permissions: Union[str, tuple[str]]):
    def decorator(method):
        @functools.wraps(method)
        def wrapper(request, *args, **kwargs):
            if request.user.is_authenticated:

                if isinstance(permissions, str):
                    perms = (permissions,)
                else:
                    perms = permissions

                if not request.user.has_perms(perms):
                    raise PermissionDenied()

                return method(request, *args, **kwargs)
            else:
                return HttpResponseRedirect(reverse('config_registration:login_redirect'))
        return wrapper
    return decorator



from django_spire.core.options.mixins import OptionsModelMixin
from django_spire.core.options.options import Options, OptionSection, Option


DEFAULT_USER_OPTIONS = Options(sections=[
    OptionSection(
        name='notification',
        options=[
            Option('email', True),
        ]
    ),
    OptionSection(
        name='system',
        options=[
            Option('timezone', 'America/Edmonton'),
        ]
    ),
])


CANADA_TIMEZONES = (
    ('Vancouver', 'Pacific Time'),
    ('Edmonton', 'Mountain Time'),
    ('Winnipeg', 'Central Time'),
    ('Toronto', 'Eastern Time'),
    ('Halifax', 'Atlantic Time'),
    ('St_Johns', 'Newfoundland Time'),
)


class UserOptionsModelMixin(OptionsModelMixin):
    _default_options = DEFAULT_USER_OPTIONS

    class Meta:
        abstract = True



# from django.shortcuts import render
# from django.shortcuts import redirect
from django.template.response import TemplateResponse

# from app.landing import forms
from django_spire.landing.context_data import (
    # contact_us_context_data,
    boneyard_context_data,
    crypt_context_data,
    asylum_context_data,
    wasteland_context_data,
    component_context_data
)
# from app.landing.utils import send_contact_us_email


def home_view(request):
    template = 'landing/page/home_page.html'
    return TemplateResponse(request, template)


# def contact_us_view(request):
#     if request.method == 'POST':
#         form = forms.ContactUsForm(request.POST)
#         if form.is_valid():
#             send_contact_us_email(**form.cleaned_data)
#             return redirect('landing:contact_form_success')
#     else:
#         form = forms.ContactUsForm()
#
#     contact_us_context_data['form'] = form
#     return TemplateResponse(request, template='landing/page/contact_us_page.html', context=contact_us_context_data)


# def contact_us_success_view(request):
#     template = 'landing/page/contact_us_success_page.html'
#     return TemplateResponse(request, template)


def boneyard_view(request):
    template = 'landing/page/boneyard.html'
    return TemplateResponse(request, template, context=boneyard_context_data)


def crypt_view(request):
    template = 'landing/page/crypt.html'
    return TemplateResponse(request, template, context=crypt_context_data)


def asylum_view(request):
    template = 'landing/page/asylum.html'
    return TemplateResponse(request, template, context=asylum_context_data)


def wasteland_view(request):
    template = 'landing/page/wasteland.html'
    return TemplateResponse(request, template, context=wasteland_context_data)



from typing import Union

from django.contrib.auth.models import User
from django.contrib.contenttypes.fields import GenericRelation
from django.db import models

from django_spire.core.comment.models import Comment


class CommentModelMixin(models.Model):
    comments = GenericRelation('core_comment.Comment', related_query_name='comment', editable=False)

    def add_comment(
            self,
            user: User,
            information: str,
            parent: Union[int, Comment, None] = None
    ):
        if isinstance(parent, Comment):
            parent = parent.pk

        return self.comments.create(information=information, user=user, parent_id=parent)

    class Meta:
        abstract = True



from django_spire.permission.models import PortalUser


def create_user(username: str, password: str) -> PortalUser:
    user = PortalUser.objects.create_user(username=username, password=password)
    user.save()
    return user


def create_super_user() -> PortalUser:
    user = PortalUser.objects.create_user(username="Super Goat", password="goatty_99")
    user.is_superuser = True
    user.is_staff = True
    user.save()
    return user



from django.contrib.auth.models import User

from django_spire.user_account.authentication.mfa.models import MfaCode


def get_or_generate_user_mfa_code(user: User) -> MfaCode:
    mfa_code = MfaCode.objects.valid_code(user)

    if mfa_code:
        return mfa_code
    else:
        return MfaCode.generate_code(user)


from django.db import models

from django_spire.history.querysets import HistoryQuerySet


class RecipeQuerySet(HistoryQuerySet):
    def by_cookbook(self, pk):
        return self.filter(cookbook_recipe__cookbook_id=pk).distinct()


class IngredientQuerySet(HistoryQuerySet):
    pass



contact_us_context_data = {
    'seo_title': 'Contact Us | Django Skeleton',
    'seo_description': 'Lorem ipsum dolor sit amet, consectetuer adipiscing elit. Aenean commodo ligula eget dolor. Aenean massa. Cum sociis natoque penatibus et magnis dis parturient.',
    'og_title': 'Contact Us | Django Skeleton',
    'og_description': 'Lorem ipsum dolor sit amet, consectetuer adipiscing elit. Aenean commodo ligula eget dolor. Aenean massa. Cum sociis natoque penatibus et magnis dis parturient.',
    'og_image': '/static/website/img/logo.png',
    'og_img_alt': 'Django Skeleton',
    'og_url': 'https://djangoskeleton.com/contact-us/'
}


boneyard_context_data = {
    'seo_title': 'Boneyard | Django Skeleton',
    'seo_description': 'Lorem ipsum dolor sit amet, consectetuer adipiscing elit. Aenean commodo ligula eget dolor. Aenean massa. Cum sociis natoque penatibus et magnis dis parturient.',
    'og_title': 'Boneyard | Django Skeleton',
    'og_description': 'Lorem ipsum dolor sit amet, consectetuer adipiscing elit. Aenean commodo ligula eget dolor. Aenean massa. Cum sociis natoque penatibus et magnis dis parturient.',
    'og_image': '/static/website/img/logo/primary/django_skeleton_primary_small.png',
    'og_img_alt': 'Django Skeleton',
    'og_url': 'https://djangoskeleton.com/boneyard/'
}


crypt_context_data = {
    'seo_title': 'Crypt | Django Skeleton',
    'seo_description': 'Lorem ipsum dolor sit amet, consectetuer adipiscing elit. Aenean commodo ligula eget dolor. Aenean massa. Cum sociis natoque penatibus et magnis dis parturient.',
    'og_title': 'Crypt | Django Skeleton',
    'og_description': 'Lorem ipsum dolor sit amet, consectetuer adipiscing elit. Aenean commodo ligula eget dolor. Aenean massa. Cum sociis natoque penatibus et magnis dis parturient.',
    'og_image': '/static/website/img/logo/primary/django_skeleton_primary_small.png',
    'og_img_alt': 'Django Skeleton',
    'og_url': 'https://djangoskeleton.com/crypt/'
}


asylum_context_data = {
    'seo_title': 'Asylum | Django Skeleton',
    'seo_description': 'Lorem ipsum dolor sit amet, consectetuer adipiscing elit. Aenean commodo ligula eget dolor. Aenean massa. Cum sociis natoque penatibus et magnis dis parturient.',
    'og_title': 'Asylum | Django Skeleton',
    'og_description': 'Lorem ipsum dolor sit amet, consectetuer adipiscing elit. Aenean commodo ligula eget dolor. Aenean massa. Cum sociis natoque penatibus et magnis dis parturient.',
    'og_image': '/static/website/img/logo/primary/django_skeleton_primary_small.png',
    'og_img_alt': 'Django Skeleton',
    'og_url': 'https://djangoskeleton.com/asylum/'
}


wasteland_context_data = {
    'seo_title': 'Wasteland | Django Skeleton',
    'seo_description': 'Lorem ipsum dolor sit amet, consectetuer adipiscing elit. Aenean commodo ligula eget dolor. Aenean massa. Cum sociis natoque penatibus et magnis dis parturient.',
    'og_title': 'Wasteland | Django Skeleton',
    'og_description': 'Lorem ipsum dolor sit amet, consectetuer adipiscing elit. Aenean commodo ligula eget dolor. Aenean massa. Cum sociis natoque penatibus et magnis dis parturient.',
    'og_image': '/static/website/img/logo/primary/django_skeleton_primary_small.png',
    'og_img_alt': 'Django Skeleton',
    'og_url': 'https://djangoskeleton.com/wasteland/'
}

component_context_data = {
    'seo_title': 'Component Library | Django Skeleton',
    'seo_description': 'Lorem ipsum dolor sit amet, consectetuer adipiscing elit. Aenean commodo ligula eget dolor. Aenean massa. Cum sociis natoque penatibus et magnis dis parturient.',
    'og_title': 'Component Library | Django Skeleton',
    'og_description': 'Lorem ipsum dolor sit amet, consectetuer adipiscing elit. Aenean commodo ligula eget dolor. Aenean massa. Cum sociis natoque penatibus et magnis dis parturient.',
    'og_image': '/static/website/img/logo/primary/django_skeleton_primary_small.png',
    'og_img_alt': 'Django Skeleton',
    'og_url': 'https://djangoskeleton.com/component/'
}



from django.apps import AppConfig


class MfaConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'app.user_account.authentication.mfa'
    label = 'authentication_mfa'



from django_spire.core.tests.test_cases import BaseTestCase
from django_spire.user_account import factories


class UserAccountFactoriesUnitTestCase(BaseTestCase):
    def test_get_or_create_user_profile(self):
        profile = factories.get_or_create_user_profile(self.super_user)
        self.assertEqual(profile.user, self.super_user)



from django.urls import path
from django_spire.user_account.views import page_views

app_name = 'user_account'
urlpatterns = [
    path("register/user/",
         page_views.register_user_form_view,
         name='register_user_form'
     ),
]



from django.urls import path
from django_spire.user_account.authentication.mfa.views import page_views

app_name = 'page'
urlpatterns = [
    path('form/',
         page_views.mfa_form_view,
         name='form'),
]



from django_spire.core.tests.test_cases import BaseTestCase

from django_spire.user_account.tests.factories import create_super_user
from django_spire.history.tests.factories import create_test_cookbook


class ActivityLogMixinTest(BaseTestCase):
    def setUp(self):
        self.test_user = create_super_user()
        self.test_recipient = create_super_user()
        self.test_subscribers = [create_super_user() for _ in range(3)]
        self.test_cookbook = create_test_cookbook()

    # TODO: Fix tests
    # def test_add_activity_no_subscribers(self):
    #     activity = self.test_cookbook.add_activity(
    #         self.test_user,
    #         'Create',
    #         'The rat created a cookbook',
    #         self.test_recipient,
    #     )
    #
    #     # Asserts
    #     self.assertTrue(True)
    #
    # def test_add_activity_with_subscribers(self):
    #     activity = self.test_cookbook.add_activity(
    #         self.test_user,
    #         'Create',
    #         'The rat created a cookbook',
    #         self.test_recipient,
    #         self.test_subscribers,
    #     )
    #
    #     # Asserts
    #     self.assertTrue(True)



from django.urls import path, include


app_name = '__init__'
urlpatterns = [
    path('', include('app.user_account.authentication.urls.admin_urls', namespace='admin')),
    path('', include('app.user_account.authentication.urls.redirect_urls', namespace='redirect')),
    path('mfa/', include('app.user_account.authentication.mfa.urls', namespace='mfa')),
]


from django import forms

from django_spire.cookbook import models


class CookbookForm(forms.ModelForm):
    linked_recipes = forms.JSONField(required=False)

    class Meta:
        model = models.Cookbook
        fields = ['name', 'description']



from django import template


register = template.Library()


@register.simple_tag(takes_context=True)
def pagination_url(context, page_number,  page_name='page',  **kwargs):
    updated_context = context.request.GET.copy()
    updated_context[page_name] = page_number
    query_string = f'?'
    for index, query_param in enumerate(updated_context):
        query_param_value = updated_context[query_param]
        if isinstance(query_param_value, str):
            # Replacing space with + sign to match how query parameters and django form filtering.
            query_param_value = query_param_value.replace(' ', '+')

        if index == 0:
            query_string = query_string + f'{query_param}={query_param_value}'
        else:
            query_string = query_string + f'&{query_param}={query_param_value}'

    return query_string


@register.simple_tag
def get_elided_page_range(page_obj, on_each_side=2, on_ends=2):
    return page_obj.paginator.get_elided_page_range(number=page_obj.number, on_each_side=on_each_side, on_ends=on_ends)



from typing import Callable

from django import forms
from django.contrib.auth.models import User
from crispy_forms.helper import FormHelper
from crispy_forms.layout import Layout, Submit


class DeleteConfirmationForm(forms.Form):
    should_delete = forms.BooleanField(required=True)

    def __init__(self, *args, obj=None, **kwargs):
        if obj is None:
            raise ValueError('Passing an object to DeleteConfirmationForm is required.')
        self.obj = obj
        super(DeleteConfirmationForm, self).__init__(*args, **kwargs)

    def save(
            self,
            user: User,
            verbs: tuple,
            delete_func: Callable = None,
            activity_func: Callable = None,
            auto_add_activity: bool = True
    ):
        if delete_func is not None:
            delete_func()
        else:
            self.obj.set_deleted()

        if activity_func is not None:
            activity_func()
        elif hasattr(self.obj, 'add_activity') and auto_add_activity:
            self.obj.add_activity(
                user=user,
                verb=verbs[1],
                information=f'{user.get_full_name()} {verbs[1].lower()} {self.obj._meta.verbose_name} "{self.obj}".'
            )




class ConfirmationForm(forms.Form):
    def __init__(self, *args, **kwargs):
        super(ConfirmationForm, self).__init__(*args, **kwargs)
        self.helper = FormHelper(self)
        self.helper.include_media = False
        self.helper.layout = Layout()
        self.helper.add_input(Submit('submit', 'Confirm', css_class='btn-success btn-sm mt-1 mb-0'))



from django.apps import AppConfig


class NutritionConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'app.cookbook.recipe.nutrition'
    label = 'cookbook_recipe_nutrition'



from django.urls import path, include

app_name = '__init__'
urlpatterns = [
    path("authentication/", include('app.user_account.authentication.urls', namespace='authentication')),
    path('profile/', include('app.user_account.profile.urls', namespace='profile')),
]



from django import template

register = template.Library()


@register.simple_tag()
def comment_form(related_obj, return_url, user, comment=None, parent=None, user_list=None):
    from django_spire.core.comment.forms import CommentForm
    return CommentForm(related_obj, return_url, user, instance=comment, parent=parent, user_list=user_list)


@register.simple_tag()
def user_list_from_content_type(related_obj):
    from django_spire.core.comment.utils import find_user_list_from_content_type
    return find_user_list_from_content_type(related_obj._meta.app_label, related_obj._meta.model_name)



from django.db import models
from django.urls import reverse

from django_spire.core.breadcrumbs import Breadcrumbs
from django_spire.history.mixins import HistoryModelMixin
from django_spire.cookbook import querysets


class Cookbook(HistoryModelMixin):
    name = models.CharField(max_length=255)
    description = models.TextField(default='')

    objects = querysets.CookbookQuerySet().as_manager()

    def __str__(self):
        return self.name

    @classmethod
    def base_breadcrumb(cls) -> Breadcrumbs:
        crumbs = Breadcrumbs()
        crumbs.add_breadcrumb('Cookbooks', reverse('cookbook:page:list'))
        return crumbs

    def breadcrumbs(self) -> Breadcrumbs:
        crumbs = Breadcrumbs()
        crumbs.add_base_breadcrumb(self._meta.model)
        if self.pk:
            crumbs.add_breadcrumb(str(self), reverse('cookbook:page:detail', kwargs={'pk': self.pk}))
        return crumbs

    class Meta:
        verbose_name = 'Cookbook'
        verbose_name_plural = 'Cookbooks'
        db_table = 'cookbook'
        permissions = [('can_approve_cookbook', 'Can Approve Cookbook')]


class CookbookRecipe(HistoryModelMixin):
    cookbook = models.ForeignKey(
        'cookbook.Cookbook',
        on_delete=models.CASCADE,
        related_name='cookbook_recipes',
        related_query_name='cookbook_recipe'
    )

    recipe = models.ForeignKey(
        'cookbook_recipe.Recipe',
        on_delete=models.CASCADE,
        related_name='cookbook_recipes',
        related_query_name='cookbook_recipe'
    )

    objects = querysets.CookbookRecipeQuerySet().as_manager()

    def __str__(self):
        return f'Cookbook Recipe {self.pk}'

    class Meta:
        verbose_name = 'Recipe'
        verbose_name_plural = 'Recipes'
        db_table = 'cookbook_recipe_m2m'



from django.urls import path, include

app_name = '__init__'
urlpatterns = [
    path('', include('app.cookbook.recipe.urls.page_urls', namespace='page')),
    path('nutrition/', include('app.cookbook.recipe.nutrition.urls', namespace='nutrition')),
]


from django.conf import settings
from django.http import HttpResponseRedirect
from django.template.response import TemplateResponse
from django.urls import reverse

from django_spire.core.views import portal_views
from django_spire.core.breadcrumbs import Breadcrumbs


def home_view(request):
    crumbs = Breadcrumbs()
    crumbs.add_breadcrumb(name='Home', href=reverse('home:home'))
    return portal_views.template_view(
        request,
        page_title='Home',
        page_description='Your Portal',
        breadcrumbs=crumbs,
        template='home/page/home_page.html'
    )


def maintenance_mode_view(request):
    if not settings.MAINTENANCE_MODE:
        return HttpResponseRedirect(request.GET.get('next', reverse('home:home')))
    else:
        return TemplateResponse(request, template='home/page/maintenance_mode_page.html')



from django.urls import path, include
from django_spire.cookbook.views import page_views


app_name = 'cookbook'
urlpatterns = [
    path('<int:pk>/delete/form', page_views.cookbook_delete_view, name='delete'),
    path('<int:pk>/detail', page_views.cookbook_detail_view, name='detail'),
    path('<int:pk>/form', page_views.cookbook_form_view, name='form'),
    path('list/', page_views.cookbook_list_view, name='list'),
]



from django.template.response import TemplateResponse

from django_glue.glue import glue_model

from django_spire.cookbook import models
from django_spire.core.shortcuts import get_object_or_null_obj
from django_spire.landing.component.utils import from_directory, from_file


def home_view(request):
    template = 'landing/page/component.html'
    return TemplateResponse(request, template)


def accordion_view(request):
    exclude_template = ['accordion']

    example = from_file('templates/landing/component/accordion/example_accordion.html')

    gallery = from_directory(
        'templates/core/accordion/',
        exclude_template=exclude_template
    )

    context = {'example': example, 'gallery': gallery}

    template = 'landing/component/accordion/accordion.html'
    return TemplateResponse(request, template, context)


def badge_view(request):
    exclude_template = ['base_badge']

    example = from_file('templates/landing/component/badge/example_badge.html')

    gallery = from_directory(
        'templates/core/badge/',
        badge_text='Sample Text',
        exclude_template=exclude_template
    )

    context = {'example': example, 'gallery': gallery}

    template = 'landing/component/badge/badge.html'
    return TemplateResponse(request, template, context)


def base_view(request):
    template = 'landing/component/base/base.html'
    return TemplateResponse(request, template)


def button_view(request):
    cookbook = get_object_or_null_obj(models.Cookbook, pk=1)
    glue_model(request, 'cookbook', cookbook)

    exclude_template = ['base_button']

    example = from_file('templates/landing/component/button/example_button.html')

    gallery = from_directory(
        'templates/core/button/',
        button_text='Sample Text',
        exclude_template=exclude_template
    )

    context = {
        'cookbook': cookbook,
        'example': example,
        'gallery': gallery,
    }

    template = 'landing/component/button/button.html'
    return TemplateResponse(request, template, context)


def card_view(request):
    exclude_template = [
        # No output/lack of context data
        'card',
        'form_card',
        'title_card'
    ]

    example = from_file('templates/landing/component/card/example_card.html')

    gallery = from_directory(
        'templates/core/card/',
        exclude_template=exclude_template
    )

    context = {'example': example, 'gallery': gallery}

    template = 'landing/component/card/card.html'
    return TemplateResponse(request, template, context)


def comment_view(request):
    # @TODO: Fix missing templates

    exclude_template = [
        # These include/extend templates that do not exist
        'comment_card',
        'comment_edit_link',
        'comment_edit_modal',
        'comment_item',
        'comment_item_ellipsis',
        'comment_modal',
        'comment_reply_element',
        'comment_reply_link',
        'comment_reply_modal',
        'comment_textarea_card',
        'comment_textarea_element',

        # No output/lack of context data
        'comment_reply_list_container',
        'comment_button_element'
    ]

    context = {}

    template = 'landing/component/comment/comment.html'
    return TemplateResponse(request, template, context)


def container_view(request):
    exclude_template = [
        # No output/lack of context data
        'container',
        'form_container'
    ]

    example = from_file('templates/landing/component/container/example_container.html')

    gallery = from_directory(
        'templates/core/container/',
        exclude_template=exclude_template
    )

    context = {'example': example, 'gallery': gallery}

    template = 'landing/component/container/container.html'
    return TemplateResponse(request, template, context)


def dropdown_view(request):
    exclude_template = [
        # No output/lack of context data
        'dropdown',
        'dropdown_link_element',
        'ellipsis_dropdown',
        'ellipsis_modal_dropdown',
        'ellipsis_dropdown_modal_link_element'
    ]

    example = from_file('templates/landing/component/dropdown/example_dropdown.html')

    gallery = from_directory(
        'templates/core/dropdown/',
        exclude_template=exclude_template
    )

    context = {'example': example, 'gallery': gallery}

    template = 'landing/component/dropdown/dropdown.html'
    return TemplateResponse(request, template, context)


def element_view(request):
    exclude_template = [
        # No output/lack of context data
        'attribute_element',
        'breadcrumb_element',
        'divider_element',
        'page_loading_element',

        'pagination_element'
    ]

    example = from_file('templates/landing/component/element/example_element.html')

    gallery = from_directory(
        'templates/core/element/',
        exclude_template=exclude_template
    )

    context = {'example': example, 'gallery': gallery}

    template = 'landing/component/element/element.html'
    return TemplateResponse(request, template, context)


def file_view(request):
    # @TODO: Fix missing templates

    exclude_template = [
        # These include/extend templates that do not exist
        'demo_detail_page',
        'file_navigation',
        'multiple_file_widget',
        'single_file_widget',
        'version_list_page',

        # No output/lack of context data
        'demo_detail_card',
        'demo_item',
        'file_item',

        # django_glue/keep-alive issue
        'demo_list_page',
    ]

    example = from_file('templates/landing/component/file/example_file.html')

    gallery = from_directory(
        'templates/core/file/',
        exclude_template=exclude_template
    )

    context = {'example': example, 'gallery': gallery}

    template = 'landing/component/file/file.html'
    return TemplateResponse(request, template, context)


def form_view(request):
    # @TODO: Fix missing templates

    exclude_template = [
        # These include/extend templates that do not exist
        'input_field',
        'search_select_form_field'
    ]

    example = from_file('templates/landing/component/form/example_form.html')

    gallery = from_directory(
        'templates/core/form/',
        exclude_template=exclude_template
    )

    context = {'example': example, 'gallery': gallery}

    template = 'landing/component/form/form.html'
    return TemplateResponse(request, template, context)


def help_view(request):
    exclude_template = ['help_button', 'help_modal']

    gallery = from_directory(
        'templates/core/help/',
        exclude_template=exclude_template
    )

    example = from_file('templates/landing/component/help/example_help.html')

    context = {'example': example, 'gallery': gallery}

    template = 'landing/component/help/help.html'
    return TemplateResponse(request, template, context)


def item_view(request):
    exclude_template = ['item', 'item_ellipsis_spacer_element']

    example = from_file('templates/landing/component/item/example_item.html')

    gallery = from_directory(
        'templates/core/item/',
        exclude_template=exclude_template
    )

    context = {'example': example, 'gallery': gallery}

    template = 'landing/component/item/item.html'
    return TemplateResponse(request, template, context)


def modal_view(request):
    exclude_template = [
        'center_modal',
        'dispatch_modal',
        'modal',
        'title_modal',
        'modal_title_content'
    ]

    example = from_file('templates/landing/component/modal/example_modal.html')

    gallery = from_directory(
        'templates/core/modal/',
        exclude_template=exclude_template
    )

    context = {'example': example, 'gallery': gallery}

    template = 'landing/component/modal/modal.html'
    return TemplateResponse(request, template, context)


def modal_content_view(request):
    template = 'landing/component/modal/modal_content.html'
    return TemplateResponse(request, template)


def navigation_view(request):
    exclude_template = [
        # Broken
        'search_bar_element',

        # No output/lack of context data
        'mobile_navigation',
        'side_navigation'
    ]

    example = from_file('templates/landing/component/navigation/example_navigation.html')

    gallery = from_directory(
        'templates/core/navigation/',
        exclude_template=exclude_template
    )

    context = {'example': example, 'gallery': gallery}

    template = 'landing/component/navigation/navigation.html'
    return TemplateResponse(request, template, context)


def notification_view(request):
    exclude_template = ['notification', 'notification_element']

    example = from_file('templates/landing/component/notification/example_notification.html')

    gallery = from_directory(
        'templates/core/notification/',
        exclude_template=exclude_template
    )

    context = {'example': example, 'gallery': gallery}

    template = 'landing/component/notification/notification.html'
    return TemplateResponse(request, template, context)


def page_view(request):
    exclude_template = [
        # No output/lack of context data
        'base_page',
        'center_card_page',
        'form_full_page',

        # django_glue/keep-alive issue
        'full_page'
    ]

    gallery = from_directory(
        'templates/core/page/',
        exclude_template=exclude_template
    )

    context = {'gallery': gallery}

    template = 'landing/component/page/_page.html'
    return TemplateResponse(request, template, context)


def tab_view(request):
    exclude_template = [
        # No output/lack of context data
        'tab',
        'tab_section_element',
        'tab_trigger_element'
    ]

    example = from_file('templates/landing/component/tab/example_tab.html')

    gallery = from_directory(
        'templates/core/tab/',
        exclude_template=exclude_template
    )

    context = {'example': example, 'gallery': gallery}

    template = 'landing/component/tab/tab.html'
    return TemplateResponse(request, template, context)



from dateutil import relativedelta
from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.utils.timezone import localtime

from django_spire.user_account.mixins import UserOptionsModelMixin
from django_spire.history.mixins import HistoryModelMixin, ActivityLogMixin
from django_spire.permission.models import PortalUser


class UserProfile(HistoryModelMixin, UserOptionsModelMixin):
    user = models.OneToOneField(
        PortalUser,
        on_delete=models.CASCADE,
        related_name='profile',
        related_query_name='profile'
    )

    mfa_valid_till_datetime = models.DateTimeField(default=localtime, editable=False)

    def __str__(self):
        return self.user.username

    def requires_mfa(self):
        return self.mfa_valid_till_datetime < localtime()

    def set_mfa_grace_period(self):
        self.mfa_valid_till_datetime = localtime() + relativedelta.relativedelta(hours=24)
        self.save()

    class Meta:
        verbose_name = 'User Profile'
        verbose_name_plural = 'User Profiles'
        db_table = 'user_account_profile'


@receiver(post_save, sender=PortalUser)
def create_or_update_user_profile(sender, instance, created, **kwargs):
    from django_spire.user_account.factories import create_user_profile

    if created:
        create_user_profile(instance)

    try:
        instance.profile.save()
    except ObjectDoesNotExist:
        create_user_profile(instance)



from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404
from django.urls import reverse

from django_spire.core.breadcrumbs import Breadcrumbs
from django_spire.core.forms import DeleteConfirmationForm, ConfirmationForm
from django_spire.core.pagination.pagination import paginate_list
from django_spire.core.redirect import safe_redirect_url
from django_spire.history.utils import add_form_activity
from django_spire.user_account import forms
from django_spire.permission.models import PortalUser
from django_spire.permission.decorators import permission_required
from django_spire.permission.permissions import generate_group_perm_data, generate_user_perm_data
from django_spire.core.views import portal_views
from django_glue.glue import glue_model
from django_glue.utils import serialize_to_json


@permission_required('permission.view_portaluser')
def user_detail_page_view(request, pk):
    user = get_object_or_404(PortalUser, pk=pk)
    group_list = user.groups.all()
    context_data = {
        'user': user,
        'group_list': group_list,
        'group_list_permission_data': [generate_group_perm_data(group) for group in group_list],
        'user_perm_data': generate_user_perm_data(user)
    }

    return portal_views.detail_view(
        request,
        context_data=context_data,
        obj=user,
        template='user_account/page/user_detail_page.html'
    )


@permission_required('permission.view_portaluser')
def user_list_page_view(request):
    user_list = PortalUser.objects.all()

    context_data = {
        'user_list': paginate_list(user_list, page_number=request.GET.get('page', 1))
    }

    return portal_views.list_view(
        request,
        context_data=context_data,
        model=PortalUser,
        template='user_account/page/user_list_page.html'
    )


@permission_required('permission.change_portaluser')
def user_form_page_view(request, pk):
    portal_user = get_object_or_404(PortalUser, pk=pk)
    glue_model(request, 'portal_user', portal_user, 'view')

    if request.method == 'POST':
        form = forms.UserForm(request.POST, instance=portal_user)
        if form.is_valid():
            portal_user = form.save()
            add_form_activity(portal_user, pk, request.user)
            return HttpResponseRedirect(reverse('user_account:profile:page:list'))
    else:
        form = forms.UserForm(instance=portal_user)

    context_data = {
        'portal_user': portal_user,
        'initial_data': serialize_to_json(form.data)
    }

    return portal_views.model_form_view(
        request,
        form=form,
        obj=portal_user,
        context_data=context_data,
        template='user_account/page/user_form_page.html'
    )


@permission_required('permission.delete_portaluser')
def user_status_form_page_view(request, pk):
    user = get_object_or_404(PortalUser, pk=pk)

    form_class = DeleteConfirmationForm if user.is_active else ConfirmationForm

    if request.method == 'POST':
        form = form_class(request.POST)
        if form.is_valid():

            user.is_active = not user.is_active
            toggle_verb = 'activated' if user.is_active else 'deactivated'
            user.save()
            user.add_activity(
                user=request.user,
                verb=toggle_verb,
                information=f'{request.user.get_full_name()} {toggle_verb} user "{user.get_full_name()}".'
            )
            return_url = safe_redirect_url(request)
            return HttpResponseRedirect(return_url)

    form = form_class()
    toggle_verb = 'Activate' if user.is_active else 'Deactivate'

    def update_breadcrumbs(breadcrumbs: Breadcrumbs):
        breadcrumbs.add_breadcrumb(name=toggle_verb)

    return portal_views.form_view(
        request,
        form=form,
        verb=toggle_verb,
        obj=user,
        breadcrumbs_func=update_breadcrumbs
    )



from django.apps import AppConfig


class UserProfileConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'app.user_account.profile'
    label = 'user_account_profile'



from django import template


register = template.Library()


@register.simple_tag
def check_permission(user, app_label, model_name, action):
    return user.has_perm(f'{app_label}.{action}_{model_name}')



from django.db import models


class RecipeCourseChoices(models.TextChoices):
    BREAKFAST = 'bre', 'Breakfast'
    APPETIZER = 'app', 'Appetizer'
    MAIN = 'mai', 'Main Course'
    DESSERT = 'des', 'Dessert'
    SOUP = 'sou', 'Soup'
    SIDE_DISH = 'sid', 'Side Dish'
    SALAD = 'sal', 'Salad'



import json

from django.core.serializers.json import DjangoJSONEncoder
from django.http import HttpResponseRedirect
from django.urls import reverse

from django_spire.core.forms import show_form_errors
from django_spire.history.utils import add_form_activity
from django_spire.user_account import forms
from django_spire.permission.models import PortalUser
from django_spire.core.views import portal_views
from django_spire.core.breadcrumbs import Breadcrumbs
from django_glue.glue import glue_model


def register_user_form_view(request):
    # Is this the best way to initialize a null object in glue?
    glue_model(request, 'portal_user', PortalUser(), 'view')

    if request.method == 'POST':
        user_form = forms.RegisterUserForm(request.POST)

        if user_form.is_valid():  # Checks to see if all forms are valid.
            user = user_form.save()  # Save each form...
            add_form_activity(user, 0, request.user)  # Add form activity. This needs to be improved.
            return HttpResponseRedirect(reverse('user_account:profile:page:list'))  # Redirect to new view.
        else:
            show_form_errors(request, user_form)
    else:
        user_form = forms.RegisterUserForm()  # If the form has an initial it will override the defaults.

    context_data = {
        # Todo: Function that takes in all of the forms and dumps the data here?
        'user_form_data': json.dumps(user_form.data, cls=DjangoJSONEncoder),
    }

    crumbs = Breadcrumbs()
    crumbs.add_breadcrumb(name='Users', href=reverse('user_account:profile:page:list'))
    crumbs.add_breadcrumb(name='Register New User')

    return portal_views.template_view(
        request,
        context_data=context_data,
        page_title='Register',
        page_description='New User',
        breadcrumbs=crumbs,
        template='user_account/page/register_user_form_page.html'
    )




from django.urls import path

from django_spire.home import views

app_name = 'home'

urlpatterns = [
    path('', views.home_view, name='home'),
    path('maintenance/mode', views.maintenance_mode_view, name='maintenance_mode'),  # Path has to match Maintenance Mode Middleware
]


from pathlib import Path

from django.template import Template, Context


def _read_file(path):
    with open(path, 'r', encoding='utf-8') as file:
        return '\n' + file.read().strip('\n')


def _render_template(html, **context):
    template = Template(html)
    context = Context(context)
    return template.render(context)


def from_directory(directory, **kwargs):
    context = {}
    directory = Path(directory)

    exclude_html = kwargs.pop('exclude_html', [])
    exclude_template = kwargs.pop('exclude_template', [])

    for path in directory.rglob('*.html'):
        file = from_file(
            path,
            exclude_html,
            exclude_template,
            **kwargs
        )

        context[path.stem] = file

    return context


def from_file(path, exclude_html = [], exclude_template = [], **kwargs):
    path = Path(path)

    html = (
        None
        if path.stem in exclude_html
        else _read_file(path)
    )

    template = (
        None
        if path.stem in exclude_template
        else _render_template(html, **kwargs)
    )

    return {
        'key': path.stem,
        'html': html,
        'template': template
    }



from django.urls import path, include
from django_spire.user_account.views import page_views

app_name = 'user_account'

urlpatterns = [
    path("register/user/",
         page_views.register_user_form_view,
         name='register_user_form'
     ),
]

urlpatterns += [
    path("authentication/", include('app.user_account.authentication.urls', namespace='authentication')),
    path("profile/", include('app.user_account.profile.urls', namespace='profile')),
]



from django.contrib import messages

def show_form_errors(request, *forms):
    for form in forms:
        for field_name, error_list in form.errors.items():
            for error in error_list.data:
                error_message = f'{field_name.title()}: {" ".join(error.messages)}'
                messages.error(request, error_message)



from django.urls import path, include

app_name = '__init__'
urlpatterns  = [
    path('', include('app.user_account.authentication.mfa.urls.page_urls', namespace='page')),
    path('', include('app.user_account.authentication.mfa.urls.redirect_urls', namespace='redirect')),
]



from django.urls import path, include
from django_spire.cookbook.recipe.views import page_views

app_name = 'recipe'
urlpatterns = [
    path('list/', page_views.recipe_list_view, name='list'),
    path('<int:pk>/detail', page_views.recipe_detail_view, name='detail'),
    path('<int:pk>/form', page_views.recipe_form_view, name='form'),
    path('<int:cookbook_pk>/recipe/<int:pk>/form', page_views.recipe_form_view, name='cookbook_form'),
    path('<int:pk>/delete/form', page_views.recipe_detail_view, name='delete'),
]



from typing import Optional, Callable

from django.http import HttpResponseRedirect
from django.template.response import TemplateResponse

from django_spire.core.forms import DeleteConfirmationForm
from django_spire.core.redirect import safe_redirect_url
from django_spire.core.forms import show_form_errors


def dispatch_modal_delete_form_content(
        request,
        *,
        obj,
        form_action: str,
        context_data: Optional[dict] = None,
        activity_func: Optional[Callable] = None,
        auto_add_activity: bool = True,
        delete_func: Optional[Callable] = None,
        verbs: tuple[str, str] = ('delete', 'deleted'),  # Present and past tense of verb
        return_url: Optional[str] = None,
        template: str = 'core/modal/content/dispatch_modal_delete_confirmation_content.html'
):
    if context_data is None:
        context_data = {}

    model_name = obj._meta.model._meta.verbose_name

    if request.method == 'POST':
        form = DeleteConfirmationForm(request.POST, obj=obj)
        if form.is_valid():
            form.save(
                user=request.user,
                verbs=verbs,
                delete_func=delete_func,
                activity_func=activity_func,
                auto_add_activity=auto_add_activity
            )

        else:
            show_form_errors(request, form)

        if return_url is None:
            return_url = safe_redirect_url(request)

        return HttpResponseRedirect(return_url)

    base_context_data = {
        'form_title': f'{model_name} - {verbs[0].title()} {obj}',
        'form_action': form_action,
        'form_description': f'Are you sure you would like to {verbs[0].lower()} {model_name} "{obj}"?',
    }

    context_data = {**base_context_data, **context_data}

    return TemplateResponse(
        request,
        template=template,
        context=context_data
    )



from django.urls import path, include


app_name = '__init__'
urlpatterns = [
    path('', include('app.cookbook.urls.page_urls', namespace='page')),
    path('recipe/', include('app.cookbook.recipe.urls', namespace='recipe')),
]



from django_spire.core.tests.test_cases import BaseTestCase
from django_spire.user_account.profile.models import UserProfile
from django_spire.user_account.tests.factories import create_user


class UserProfileTestCase(BaseTestCase):
    def setUp(self):
        self.user = create_user(username="Wesley", password="goat_99")

    def test_user_profile_receiver(self):
        profile = self.user.profile
        self.assertTrue(UserProfile.objects.filter(user=profile.user_id).exists())



from functools import wraps

from django import template


register = template.Library()


DIRECTIVES = (
    'x-bind',
    'x-cloak',
    'x-data',
    'x-effect',
    'x-for',
    'x-html',
    'x-id',
    'x-if',
    'x-ignore',
    'x-init',
    'x-model',
    'x-modelable',
    'x-on',
    'x-ref',
    'x-show',
    'x-teleport',
    'x-text',
    'x-transition'
)


MAGIC = (
    '$data',
    '$dispatch',
    '$el',
    '$id',
    '$nextTick',
    '$refs',
    '$root',
    '$store',
    '$watch'
)


@register.filter(name='lowercase')
def lowercase(value):
    return value.lower()


@register.inclusion_tag('landing/component/subsection/subheader.html')
def title(title="Subheading"):
    return {'title': title}


def alpine(func):
    @wraps(func)
    def wrapper(context, *args, **kwargs):
        attributes = {}

        keys = kwargs.keys()

        for key in list(keys):
            if key in DIRECTIVES or MAGIC:
                attributes[key] = kwargs.pop(key)

        context.update(kwargs)
        context.update(attributes)

        return func(context, *args, **kwargs)

    return wrapper


@register.inclusion_tag('core/button/brayden_base.html', takes_context=True)
@alpine
def button(context, **kwargs):
    return context



from django.urls import path
from django_spire.user_account.authentication.mfa.views import redirect_views

app_name = 'redirect'
urlpatterns = [
    path('notification/',
         redirect_views.mfa_notification_redirect_view,
         name='notification'),
]



# def subject_verbose(inquiry):
#     from app.landing.forms import SUBJECT_TOPIC_CHOICES
#     return dict(SUBJECT_TOPIC_CHOICES)[inquiry]
#
#
# def send_contact_us_email(**kwargs):
#     from django.core.mail import EmailMultiAlternatives
#     from django.template.loader import get_template
#
#     text_content = f'''
#         New message from {kwargs['name']} ({kwargs['email']}).
#         {kwargs['company']}
#         {kwargs['message']}
#     '''
#
#     subject = 'Company Name - New Inquiry'
#     html = get_template('core/email/contact_us_email_template.html')
#
#     context_data = {
#         'name': kwargs['name'], 'company': kwargs['company'], 'email': kwargs['email'],
#         'inquiry': kwargs['inquiry'], 'message': kwargs['message']
#     }
#
#     subject, from_email, to = subject, 'companyname@example.com', 'example@example.com'
#     html_content = html.render(context_data)
#     msg = EmailMultiAlternatives(subject, text_content, from_email, [to])
#     msg.attach_alternative(html_content, "text/html")
#     msg.send()



from django_spire.cookbook.recipe import models


def update_or_create_recipe_ingredients(recipe, ingredient_data: list[dict]) -> list[models.Ingredient]:
    previous_ingredients = recipe.ingredients.active().values_list('id', flat=True)

    ingredients = []
    for raw_ingredient in ingredient_data:
        ingredient = models.Ingredient(**raw_ingredient)
        ingredient.recipe = recipe
        ingredient.save()
        ingredients.append(ingredient)

    deleted_ids = set(previous_ingredients) - set([ingredient.id for ingredient in ingredients])

    for deleted_ingredient in models.Ingredient.objects.filter(id__in=deleted_ids):
        deleted_ingredient.set_deleted()

    return ingredients



from django.urls import path
from django_spire.user_account.authentication.views import redirect_views

app_name = 'authentication'
urlpatterns = [
    path('login/redirect/',
         redirect_views.login_redirect_view,
         name='login'),

    path('logout/redirect/',
         redirect_views.logout_redirect_view,
         name='logout'),
]



from django_spire.permission.models import PortalGroup


def bulk_create_groups_from_names(names: list[str]):
    existing_groups = PortalGroup.objects.all().values_list('name', flat=True)
    new_groups = [PortalGroup(name=name) for name in names if name not in existing_groups]
    return PortalGroup.objects.bulk_create(new_groups)



from typing import Union

from django_spire.core.notification.models import Notification


def create_notification(
        email: str,
        title: str,
        body: str,
        url: Union[str, None] = None,
        name: Union[str, None] = None
):
    return Notification.objects.create(
        email=email,
        name=name,
        title=title,
        body=body,
        url=url
    )



from __future__ import annotations

import json

from django import forms

from django_spire.cookbook.recipe import models, factories


class IngredientsField(forms.CharField):
    def clean(self, value: str) -> list[dict]:
        cleaned_data = []
        if value:
            value = json.loads(value)
            for ingredient in value:
                form = IngredientForm(ingredient)
                if form.is_valid():
                    cleaned_data.append(form.cleaned_data)
                else:
                    raise forms.ValidationError(form.errors)

        return cleaned_data


class RecipeForm(forms.ModelForm):
    ingredients = IngredientsField()

    def save(self, commit=True):
        recipe = super().save(commit=commit)
        factories.update_or_create_recipe_ingredients(recipe, self.cleaned_data['ingredients'])
        return recipe

    class Meta:
        model = models.Recipe
        exclude = []


class IngredientForm(forms.ModelForm):
    id = forms.IntegerField(required=False)

    class Meta:
        model = models.Ingredient
        exclude = ['recipe']



from django.core.paginator import Paginator, Page


def paginate_list(object_list: list, page_number: int = 1, per_page: int = 50) -> Page:
    paginator = Paginator(object_list, per_page)
    return paginator.get_page(page_number)



# from captcha.fields import ReCaptchaField
# from captcha.widgets import ReCaptchaV2Checkbox
# from django import forms
#
# from crispy_forms.helper import FormHelper
# from crispy_forms.layout import Layout, Row, Column, HTML
#
#
# SUBJECT_TOPIC_CHOICES = (
#     ('aaa', 'Example_1'),
#     ('bbb', 'Example_2'),
#     ('ccc', 'Example_3'),
#     ('ddd', 'Example_4'),
#     ('oth', 'Other'),
# )
#
#
# class ContactUsForm(forms.Form):
#     def __init__(self, *args, **kwargs):
#         super(ContactUsForm, self).__init__(*args, **kwargs)
#
#         self.fields['name'] = forms.CharField(max_length=124)
#         self.fields['company'] = forms.CharField(max_length=124)
#         self.fields['email'] = forms.EmailField()
#         self.fields['inquiry'] = forms.ChoiceField(choices=SUBJECT_TOPIC_CHOICES)
#         self.fields['message'] = forms.CharField(widget=forms.Textarea(attrs={'rows': 4}))
#         self.fields['captcha'] = ReCaptchaField(widget=ReCaptchaV2Checkbox)
#
#         self.helper = FormHelper(self)
#         self.helper.include_media = False
#
#         self.helper.layout = Layout(
#             Row(
#                 Column('name', css_class='form-group col-12'),
#                 Column('company', css_class='form-group col-12'),
#                 Column('email', css_class='form-group col-12'),
#                 Column('inquiry', css_class='form-group col-12'),
#                 Column('message', css_class='form-group col-12'),
#             ),
#             Row(
#                 Column('captcha', css_class='form-group col-12'),
#             ),
#
#             HTML("{% include 'core/element/form_button_element.html' %}"),
#         )



from typing import Type
from django_spire.history.mixins import HistoryModelMixin
from django.contrib.auth.models import User


def add_form_activity(obj: Type[HistoryModelMixin], pk: int, user: User,):
    verb = 'updated' if pk != 0 else 'created'
    obj.add_activity(
        user=user,
        verb=verb,
        information=f'{user.get_full_name()} {verb} {obj._meta.verbose_name} "{obj}".'
    )



from django.shortcuts import get_object_or_404, redirect
from django.urls import reverse

from django_spire.core.forms import show_form_errors
from django_spire.history.utils import add_form_activity
from django_glue.glue import glue_model, glue_query_set

from django_spire.cookbook.recipe.models import Recipe
from django_spire.core.views import portal_views
from django_spire.cookbook import models, forms, factories

from django_spire.core.shortcuts import get_object_or_null_obj
from django_spire.permission.decorators import permission_required


def cookbook_delete_view(request, pk):
    cookbook = get_object_or_404(models.Cookbook, pk=pk)
    return portal_views.delete_form_view(
        request,
        obj=cookbook,
        return_url=request.GET.get('return_url', reverse('cookbook:page:list')),
    )


@permission_required('perms.cookbook.view_cookbook')
def cookbook_detail_view(request, pk):
    cookbook = get_object_or_404(models.Cookbook, pk=pk)
    context_data = {
        'cookbook': cookbook,
        'recipes': Recipe.objects.by_cookbook(pk)
    }

    return portal_views.detail_view(
        request,
        obj=cookbook,
        context_data=context_data,
        template='cookbook/page/cookbook_detail_page.html'
    )


def cookbook_form_view(request, pk):
    cookbook = get_object_or_null_obj(models.Cookbook, pk=pk)

    glue_model(request, 'cookbook', cookbook, 'view')
    glue_query_set(request,  'recipe_queryset', Recipe.objects.active(), 'view')

    if request.method == 'POST':
        form = forms.CookbookForm(request.POST, instance=cookbook)
        if form.is_valid():
            cookbook = form.save()
            factories.link_cookbook_recipes(cookbook, form.cleaned_data['linked_recipes'])
            add_form_activity(cookbook, pk, request.user)
            return redirect(request.GET.get('return_url', reverse('cookbook:page:list')))
        else:
            show_form_errors(request, form)
    else:
        form = forms.CookbookForm(instance=cookbook)

    return portal_views.form_view(
        request,
        form=form,
        obj=cookbook,
        template='cookbook/page/cookbook_form_page.html'
    )

@permission_required('perms.cookbook.view_cookbook')
def cookbook_list_view(request):
    context_data = {
        'cookbooks': models.Cookbook.objects.all()
    }

    return portal_views.list_view(
        request,
        model=models.Cookbook,
        context_data=context_data,
        template='cookbook/page/cookbook_list_page.html'
    )



from django.apps import AppConfig


class AuthenticationConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'app.user_account.authentication'
    label = 'user_account_authentication'



import json
from typing import Optional

from django.contrib.contenttypes.models import ContentType
from django.db.models import Model


def get_object_or_null_obj(model, pk: int, **kwargs):
    try:
        return model.objects.get(pk=pk, **kwargs)
    except model.DoesNotExist:
        return model()


def get_object_or_none(model, pk, **kwargs):
    try:
        return model.objects.get(pk=pk, **kwargs)
    except model.DoesNotExist:
        return None


def process_request_body(request):
    body_unicode = request.body.decode('utf-8')
    return json.loads(body_unicode)['data']


def model_object_from_app_label(app_label, model_name, object_pk) -> Optional[Model]:
    try:
        content_type = ContentType.objects.get(app_label=app_label, model=model_name)
    except ContentType.DoesNotExist:
        return None

    model_class = content_type.model_class()

    return get_object_or_none(model_class, pk=object_pk)



from django.contrib.auth.models import User

from django_spire.user_account.profile.models import UserProfile

from django_spire.permission.models import PortalUser


def get_or_create_user_profile(user: User) -> UserProfile:
    try:
        return user.profile
    except UserProfile.DoesNotExist:
        return create_user_profile(user)


def create_user_profile(user: User) -> UserProfile:
    return UserProfile.objects.create(user=user)


def register_new_user(
        first_name: str,
        last_name: str,
        email: str,
        password: str
):
    return PortalUser.objects.create_user(
        username=email.lower(),
        email=email.lower(),
        password=password,
        first_name=first_name,
        last_name=last_name
    )



from django_spire.cookbook.models import Cookbook, CookbookRecipe


def link_cookbook_recipes(cookbook: Cookbook, recipe_ids: list[id]) -> list[CookbookRecipe]:
    previous_recipes = cookbook.cookbook_recipes.active().values_list('recipe_id', flat=True)

    cookbook_recipes = []
    for recipe_id in set(recipe_ids) - set(previous_recipes):
        cookbook_recipe = cookbook.cookbook_recipes.create(recipe_id=recipe_id)
        cookbook_recipes.append(cookbook_recipe)

    deleted_ids = set(previous_recipes) - set(recipe_ids)
    for delete_cookbook_recipes in cookbook.cookbook_recipes.filter(recipe_id__in=deleted_ids):
        delete_cookbook_recipes.set_deleted()

    return cookbook_recipes



from typing import Optional

from django.shortcuts import get_object_or_404, redirect
from django.urls import reverse

from django_spire.core.forms import show_form_errors
from django_spire.core.shortcuts import get_object_or_null_obj
from django_spire.history.utils import add_form_activity
from django_glue.glue import glue_model, glue_query_set

from django_spire.cookbook.models import Cookbook
from django_spire.cookbook.recipe import models, forms
from django_spire.cookbook.recipe.nutrition.models import NutritionFact
from django_spire.core.views import portal_views


def recipe_delete_view(request, pk):
    recipe = get_object_or_404(models.Recipe, pk=pk)
    return portal_views.delete_form_view(
        request,
        obj=recipe,
        return_url=request.GET.get('return_url', reverse('recipe_list')),
    )


def recipe_detail_view(request, pk):
    recipe = get_object_or_404(models.Recipe, pk=pk)
    context_data = {
        'recipe': recipe,
        'ingredients': recipe.ingredients.active()
    }
    return portal_views.detail_view(
        request,
        obj=recipe,
        context_data=context_data,
        template='cookbook/recipe/page/recipe_detail_page.html'
    )


def recipe_form_view(request, pk, cookbook_pk: Optional[int]= None):
    if cookbook_pk is not None:
        cookbook = get_object_or_404(Cookbook, pk=cookbook_pk)

    recipe = get_object_or_null_obj(models.Recipe, pk=pk)

    if recipe.pk is None:
        ingredients = models.Ingredient.objects.none()
        nutrition_facts = NutritionFact.objects.none()
    else:
        ingredients = recipe.ingredients.active()
        nutrition_facts = recipe.nutrition_facts.active()

    glue_model(request, 'recipe', recipe, 'view')

    glue_query_set(request, 'ingredient_queryset', ingredients, 'view', exclude=['recipe'])
    glue_query_set(request, 'nutrition_facts_queryset', nutrition_facts, 'view', exclude=['recipe'])
    glue_model(request, 'ingredient_template', models.Ingredient(), 'view', exclude=['recipe'])
    glue_model(request, 'nutrition_fact_template', NutritionFact(), 'view', exclude=['recipe'])

    if request.method == 'POST':
        form = forms.RecipeForm(request.POST, instance=recipe)
        if form.is_valid():
            recipe = form.save()

            if cookbook_pk is not None:
                recipe.cookbook_recipes.create(cookbook=cookbook)

            add_form_activity(recipe, pk, request.user)
            return redirect(request.GET.get('return_url', reverse('cookbook:recipe:page:detail', kwargs={'pk': recipe.pk})))
        else:
            show_form_errors(request, form)
    else:
        form = forms.RecipeForm(instance=recipe)

    return portal_views.form_view(
        request,
        form=form,
        obj=recipe,
        template='cookbook/recipe/page/recipe_form_page.html'
    )


def recipe_list_view(request):
    context_data = {
        'recipes': models.Recipe.objects.active()
    }

    return portal_views.list_view(
        request,
        model=models.Recipe,
        context_data=context_data,
        template='cookbook/recipe/page/recipe_list_page.html'
    )



from django.urls import path

from django_spire.landing.component import views


app_name = 'component'


urlpatterns = [
    path('', views.home_view, name='home'),

    path('accordion', views.accordion_view, name='accordion'),
    path('badge', views.badge_view, name='badge'),
    path('base', views.base_view, name='base'),
    path('button', views.button_view, name='button'),
    path('card', views.card_view, name='card'),
    path('comment', views.comment_view, name='comment'),
    path('container', views.container_view, name='container'),
    path('dropdown', views.dropdown_view, name='dropdown'),
    path('element', views.element_view, name='element'),
    path('file', views.file_view, name='file'),
    path('form', views.form_view, name='form'),
    path('help', views.help_view, name='help'),
    path('item', views.item_view, name='item'),
    path('modal', views.modal_view, name='modal'),
    path('modal_content', views.modal_content_view, name='modal_content'),
    path('navigation', views.navigation_view, name='navigation'),
    path('notification', views.notification_view, name='notification'),
    path('page', views.page_view, name='page'),
    path('tab', views.tab_view, name='tab')
]



from django.db import models
from django.urls import reverse

from django_spire.core.breadcrumbs import Breadcrumbs
from django_spire.history.mixins import HistoryModelMixin
from django_spire.cookbook.recipe import querysets
from django_spire.cookbook.recipe.choices import RecipeCourseChoices


class Recipe(HistoryModelMixin):
    name = models.CharField(max_length=255)
    description = models.TextField(default='')

    course = models.CharField(
        max_length=3,
        choices=RecipeCourseChoices.choices,
        default=RecipeCourseChoices.MAIN
    )

    prep_time = models.IntegerField(default=15)
    cook_time = models.IntegerField(default=30)

    servings = models.IntegerField(default=1)

    objects = querysets.RecipeQuerySet.as_manager()

    def __str__(self):
        return self.name

    def breadcrumbs(self):
        crumbs = Breadcrumbs()
        if self.pk:
            crumbs.add_breadcrumb(str(self), reverse('cookbook:recipe:page:detail', kwargs={'pk': self.pk}))
        return crumbs

    def choice_verbose(self):
        return RecipeCourseChoices(self.course).label

    def total_time(self) -> int:
        return self.prep_time + self.cook_time

    def set_deleted(self):
        for cookbook_recipe in self.cookbook_recipes.active():
            cookbook_recipe.set_deleted()
        super().set_deleted()

    class Meta:
        verbose_name = 'Recipe'
        verbose_name_plural = 'Recipes'
        db_table = 'cookbook_recipe'


class Ingredient(HistoryModelMixin):
    recipe = models.ForeignKey(
        'cookbook_recipe.Recipe',
        on_delete=models.CASCADE,
        related_name='ingredients',
        related_query_name='ingredient'
    )
    name = models.CharField(max_length=255)
    quantity = models.CharField(max_length=255, default='')
    directions = models.TextField(default='')

    objects = querysets.IngredientQuerySet.as_manager()

    def __str__(self):
        return self.name

    class Meta:
        verbose_name = 'Ingredient'
        verbose_name_plural = 'Ingredients'
        db_table = 'cookbook_ingredient'



from abc import ABC
from typing import Union

from django.conf import settings
from django.core.mail import EmailMessage


class EmailHelper(ABC):
    def __init__(
            self,
            to: Union[list, str],
            cc: Union[list, None] = None,
            bcc: Union[list, None] = None,
            fail_silently: bool = False
    ):
        if isinstance(to, str):
            self.to = [to]
        else:
            self.to = to

        self.cc = cc
        self.bcc = bcc
        self.from_email = settings.DEFAULT_FROM_EMAIL
        self.fail_silently = fail_silently


class SendGridEmailHelper(EmailHelper):
    def __init__(
            self,
            to: Union[list, str],
            template_data: dict,
            template_id: str = settings.SENDGRID_TEMPLATE_ID,
            cc: Union[list, None] = None,
            bcc: Union[list, None] = None,
            fail_silently: bool = False
    ):
        super(EmailHelper, self).__init__(to, cc, bcc, fail_silently)
        self.template_id = template_id
        self.template_data = template_data

    def send(self):
        msg = EmailMessage(
            from_email=self.from_email,
            to=self.to,
            cc=self.cc,
            bcc=self.bcc
        )
        msg.template_id = self.template_id
        msg.dynamic_template_data = self.template_data
        msg.send(fail_silently=self.fail_silently)
